﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Reflection;
using UnityEditor;
using UnityEngine;
using UnityEngine.UI;

public static class ShortcutExtensions
{
    /* Key mapping
     * % - Ctrl / Cmd
     * # - Shift
     * & - Alt
     * LEFT/RIGHT/UP/DOWN - Arrow keys
     * F1...F2 - F keys
     * HOME, END, PGUP, PGDN
     */

    /* Rules:
     * Create           <=>         ctrl + alt
     * Edit             <=>         ctrl / ctrl + shift
     * GUI              <=>         ctrl + alt
     * Hierarchy        <=>         ctrl + shift
     */

    #region Create

    [MenuItem("Shortcuts/Create/Empty child %&n", priority = 1)] // ctrl + alt + n
    private static void CreateEmptyChild()
    {
        GameObject[] selected = Selection.objects.Select(x => x as GameObject).ToArray();
        List<GameObject> created = new List<GameObject>();
        foreach(GameObject go in selected)
        {
            GameObject child = new GameObject("GameObject");
            Undo.RegisterCreatedObjectUndo(child, "Add child");
            child.transform.SetParent(go.transform);
            if(child.transform.parent.GetComponent<RectTransform>() != null)
                child.AddComponent<RectTransform>();
            child.transform.localPosition = Vector3.zero;
            child.transform.localRotation = Quaternion.identity;
            child.transform.localScale = Vector3.one;
            created.Add(child);
        }
        Selection.objects = created.ToArray();
    }

    [MenuItem("Shortcuts/Create/Empty child %&n", true)]
    private static bool CreateEmptyChildValidation()
    {
        return Selection.gameObjects.Length != 0;
    }

    #endregion

    #region Edit

    [MenuItem("Shortcuts/Edit/Paste as child %#v", priority = 1)] // ctrl + shift + v
    private static void PasteAsChildren()
    {
        GameObject[] selected = Selection.gameObjects;
        List<GameObject> pasted = new List<GameObject>();
        foreach(GameObject go in selected)
        {
            GameObject child = new GameObject("GameObject");
            Undo.RegisterCreatedObjectUndo(child, "Paste as child");
            child.transform.SetParent(go.transform);
            Selection.activeGameObject = child;
            EditorApplication.ExecuteMenuItem("Edit/Paste");
            pasted.Add(Selection.activeGameObject);
            UnityEngine.Object.DestroyImmediate(child);
        }
        Selection.objects = pasted.ToArray();
    }

    [MenuItem("Shortcuts/Edit/Paste as child %#v", true)]
    private static bool PasteAsChildrenValidation()
    {
        return Selection.gameObjects.Length != 0;
    }

    [MenuItem("Shortcuts/Edit/Pause game &q", priority = 51)] // ctrl + q
    private static void PauseGame()
    {
        EditorApplication.isPaused = !EditorApplication.isPaused;
    }

    [MenuItem("Shortcuts/Edit/Pause game &q", true)]
    private static bool PauseGameValidation()
    {
        return EditorApplication.isPlaying;
    }

    #endregion

    #region GUI

    #region GUI/Canvas

    private static Vector2 CanvasHDSize = new Vector2(1920, 1080);

    [MenuItem("Shortcuts/GUI/Canvas/Add canvas pack %&c", priority = 1)] // ctrl + alt + c
    private static void CanvasCreatePack()
    {
        GameObject[] selected = Selection.gameObjects;

        foreach(GameObject go in selected)
        {
            if(go.GetComponent<Canvas>() == null)
                Undo.AddComponent<Canvas>(go);
            if(go.GetComponent<CanvasScaler>() == null)
                Undo.AddComponent<CanvasScaler>(go);
            if(go.GetComponent<GraphicRaycaster>() == null)
                Undo.AddComponent<GraphicRaycaster>(go);
            if(go.GetComponent<CanvasGroup>() == null)
                Undo.AddComponent<CanvasGroup>(go);
        }
    }

    [MenuItem("Shortcuts/GUI/Canvas/Add canvas pack %&c", true)]
    private static bool CanvasCreatePackValidation()
    {
        return Selection.gameObjects.Length != 0;
    }

    [MenuItem("Shortcuts/GUI/Canvas/Change canvas into HD world space %&h", priority = 2)] // ctrl + alt + h
    private static void CanvasWorldSpaceHD()
    {
        GameObject[] selected = Selection.gameObjects;

        Undo.RegisterCompleteObjectUndo(selected, "Change canvas properties");

        foreach(GameObject go in selected)
        {
            go.GetComponent<Canvas>().renderMode = RenderMode.WorldSpace;
            go.GetComponent<RectTransform>().sizeDelta = new Vector2(CanvasHDSize.x, CanvasHDSize.y);
            go.GetComponent<RectTransform>().localScale = new Vector3(1f / CanvasHDSize.x, 1f / CanvasHDSize.x, 1f);
        }
    }

    [MenuItem("Shortcuts/GUI/Canvas/Change canvas into HD world space %&h", true)]
    private static bool CanvasWorldSpaceHDValidation()
    {
        return Selection.gameObjects.Length != 0 && Selection.gameObjects.All(x => x.GetComponent<Canvas>() != null);
    }

    [MenuItem("Shortcuts/GUI/Canvas/Change canvas into HD overlay %&o", priority = 3)] // ctrl + alt + o
    private static void CanvasCameraOverlay()
    {
        GameObject[] selected = Selection.gameObjects;

        Undo.RegisterCompleteObjectUndo(selected, "Change canvas properties");

        foreach(GameObject go in selected)
        {
            go.GetComponent<Canvas>().renderMode = RenderMode.ScreenSpaceOverlay;
            go.GetComponent<CanvasScaler>().uiScaleMode = CanvasScaler.ScaleMode.ScaleWithScreenSize;
            go.GetComponent<CanvasScaler>().screenMatchMode = CanvasScaler.ScreenMatchMode.MatchWidthOrHeight;
            go.GetComponent<CanvasScaler>().referenceResolution = CanvasHDSize;
            go.GetComponent<CanvasScaler>().matchWidthOrHeight = 1f;
        }
    }

    [MenuItem("Shortcuts/GUI/Canvas/Change canvas into HD overlay %&o", true)]
    private static bool CanvasCameraOverlayValidation()
    {
        return Selection.gameObjects.Length != 0 && Selection.gameObjects.All(x => x.GetComponent<Canvas>() != null && x.GetComponent<CanvasScaler>() != null);
    }

    #endregion

    [MenuItem("Shortcuts/GUI/Image %&i", priority = 51)] // ctrl + alt + i
    private static void AddImageComponent()
    {
        foreach(GameObject go in Selection.gameObjects)
        {
            if(go.GetComponent<Image>() == null)
                Undo.AddComponent<Image>(go);
        }
    }

    [MenuItem("Shortcuts/GUI/Image %&i", true)]
    private static bool AddImageComponentValidation()
    {
        return Selection.gameObjects.Length != 0 && Selection.gameObjects.Any(x => x.GetComponent<Image>() == null);
    }

    [MenuItem("Shortcuts/GUI/Button %&b", priority = 52)] // ctrl + alt + b
    private static void AddButtonComponent()
    {
        foreach(GameObject go in Selection.gameObjects)
        {
            if(go.GetComponent<Image>() == null)
                Undo.AddComponent<Image>(go);
            if(go.GetComponent<Button>() == null)
                Undo.AddComponent<Button>(go);
        }
    }

    [MenuItem("Shortcuts/GUI/Button %&b", true)]
    private static bool AddButtonComponentValidation()
    {
        return Selection.gameObjects.Length != 0 && Selection.gameObjects.Any(x => x.GetComponent<Image>() == null && x.GetComponent<Button>() == null);
    }

    [MenuItem("Shortcuts/GUI/Text %&t", priority = 53)] // ctrl + alt + t
    private static void AddTextComponent()
    {
        foreach(GameObject go in Selection.gameObjects)
        {
            if(go.GetComponent<Text>() == null)
                Undo.AddComponent<Text>(go);
        }
    }

    [MenuItem("Shortcuts/GUI/Text %&t", true)]
    private static bool AddTextComponentValidation()
    {
        return Selection.gameObjects.Length != 0 && Selection.gameObjects.Any(x => x.GetComponent<Text>() == null);
    }

    [MenuItem("Shortcuts/GUI/Mask %&m", priority = 54)] // ctrl + alt + m
    private static void AddMaskComponent()
    {
        foreach(GameObject go in Selection.gameObjects)
        {
            if(go.GetComponent<Mask>() == null)
                Undo.AddComponent<Mask>(go);
            if(go.GetComponent<Image>() == null)
                Undo.AddComponent<Image>(go);
        }
    }

    [MenuItem("Shortcuts/GUI/Mask %&m", true)]
    private static bool AddMaskComponentValidation()
    {
        return Selection.gameObjects.Length != 0 && Selection.gameObjects.Any(x => x.GetComponent<Mask>() == null);
    }

    #region Layout

    [MenuItem("Shortcuts/GUI/LayoutElement %&l", priority = 101)] // ctrl + alt + l
    private static void AddLayoutElementComponent()
    {
        foreach(GameObject go in Selection.gameObjects)
        {
            if(go.GetComponent<LayoutElement>() == null)
                Undo.AddComponent<LayoutElement>(go);
        }
    }

    [MenuItem("Shortcuts/GUI/LayoutElement %&l", true)]
    private static bool AddLayoutElementComponentValidation()
    {
        return Selection.gameObjects.Length != 0 && Selection.gameObjects.Any(x => x.GetComponent<LayoutElement>() == null);
    }

    [MenuItem("Shortcuts/GUI/ContentSizeFitter %&w", priority = 102)] // ctrl + alt + w
    private static void AddContentSizeFitterComponent()
    {
        foreach(GameObject go in Selection.gameObjects)
        {
            if(go.GetComponent<ContentSizeFitter>() == null)
                Undo.AddComponent<ContentSizeFitter>(go);
        }
    }

    [MenuItem("Shortcuts/GUI/ContentSizeFitter %&w", true)]
    private static bool AddContentSizeFitterComponentValidation()
    {
        return Selection.gameObjects.Length != 0 && Selection.gameObjects.Any(x => x.GetComponent<ContentSizeFitter>() == null);
    }

    #endregion

    #endregion

    #region Hierarchy

    [MenuItem("Shortcuts/Hierarchy/Move object up %#UP", priority = 1)] // ctrl + shift + arrow up
    private static void HierarchyMoveObjectUp()
    {
        Transform[] selected = Selection.transforms;

        Undo.RegisterFullObjectHierarchyUndo(selected[0].parent, "Move object");

        foreach(Transform t in selected)
        {
            t.SetSiblingIndex(t.GetSiblingIndex() - 1);
        }
    }

    [MenuItem("Shortcuts/Hierarchy/Move object up %#UP", true)]
    private static bool HierarchyMoveObjectUpValidation()
    {
        if(Selection.transforms.Length == 0
            || Selection.transforms.Any(x => x.parent == null || x.parent != Selection.transforms[0].parent || x.GetSiblingIndex() == 0))
            return false;

        return true;
    }

    [MenuItem("Shortcuts/Hierarchy/Move object down %#DOWN", priority = 2)] // ctrl + shift + arrow down
    private static void HierarchyMoveObjectDown()
    {
        Transform[] selected = Selection.transforms;

        Undo.RegisterFullObjectHierarchyUndo(selected[0].parent, "Move object");

        foreach(Transform t in selected.Reverse())
        {
            t.SetSiblingIndex(t.GetSiblingIndex() + 1);
        }
    }

    [MenuItem("Shortcuts/Hierarchy/Move object down %#DOWN", true)]
    private static bool HierarchyMoveObjectDownValidation()
    {
        if(Selection.transforms.Length == 0
            || Selection.transforms.Any(x => x.parent == null || x.parent != Selection.transforms[0].parent || x.GetSiblingIndex() == x.parent.childCount - 1))
            return false;

        return true;
    }

    #endregion

    #region MonoBehaviour extensions

    [MenuItem("CONTEXT/Component/Separate component as child", priority = 1001)]
    static void SeparateComponentAsChild(MenuCommand command)
    {
        Component body = (Component)command.context;
        GameObject child = new GameObject("Child: " + body.GetType().Name);
        Undo.RegisterCreatedObjectUndo(child, "Add child");
        child.transform.SetParent(body.transform);
        Component c = Undo.AddComponent(child, body.GetType());
        body.CopyTo(ref c);
        Undo.DestroyObjectImmediate(body);
        Selection.activeGameObject = child;
    }

    [MenuItem("CONTEXT/Component/Separate component as child", true)]
    static bool SeparateComponentAsChildValidation(MenuCommand command)
    {
        Component body = (Component)command.context;
        return !(body is Transform);
    }

    #endregion

    public static void CopyTo<T>(this T copyFrom, ref Component copyTo) where T : Component
    {
        Type type = copyTo.GetType();
        if (type != copyFrom.GetType()) return; // type mis-match
        BindingFlags flags = BindingFlags.Public | BindingFlags.NonPublic | BindingFlags.Instance | BindingFlags.Default | BindingFlags.DeclaredOnly;
        PropertyInfo[] pinfos = type.GetProperties(flags);
        foreach (var pinfo in pinfos)
        {
            if (pinfo.CanWrite)
            {
                try
                {
                    pinfo.SetValue(copyTo, pinfo.GetValue(copyFrom, null), null);
                }
                catch { } // In case of NotImplementedException being thrown. For some reason specifying that exception didn't seem to catch it, so I didn't catch anything specific.
            }
        }
        FieldInfo[] finfos = type.GetFields(flags);
        foreach (var finfo in finfos)
        {
            finfo.SetValue(copyTo, finfo.GetValue(copyFrom));
        }
        return;
    }
}
